/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>

#include <osip_msg.h>

extern psp_core_t *core;


char *
to_tag_new_random ()
{
  char *tmp = (char *) smalloc (33);
  unsigned int number = new_random_number ();

  sprintf (tmp, "%u", number);
  return tmp;
}

char *
from_tag_new_random ()
{
  return to_tag_new_random ();
}

/* status code between 101 and 299 are forbidden...
   Probably only answer code>.
*/
int
osip_msg_sfp_build_response (sip_t ** dest, int status, sip_t * request)
{
  return osip_msg_uap_build_response (dest, status, request);
}

/* status code between 101 and 299 are forbidden...
   (except for REGISTER where 200 is allowed).
*/
int
osip_msg_uap_build_response (sip_t ** dest, int status, sip_t * request)
{
  generic_param_t *tag;
  char *statuscode;
  sip_t *response;
  char *tmp;
  int pos;
  int i;

  *dest = NULL;
  i = msg_init (&response);
  if (i != 0)
    return -1;

  msg_setversion (response, sgetcopy ("SIP/2.0"));
  statuscode = (char *) smalloc (5 * sizeof (char));
  sprintf (statuscode, "%i", status);
  msg_setstatuscode (response, statuscode);

  tmp = msg_getreason (status);
  if (tmp == NULL)
    msg_setreasonphrase (response, sgetcopy ("Unknown status code"));
  else
    msg_setreasonphrase (response, tmp);

  msg_setmethod (response, NULL);
  msg_seturi (response, NULL);

  i = to_clone (request->to, &(response->to));
  if (i != 0)
    goto mcubr_error_1;

  i = to_get_tag (response->to, &tag);
  if (i != 0)
    {				/* we only add a tag if it does not already contains one! */
      if (status == 200 && MSG_IS_REGISTER (request))
	{
	  to_set_tag (response->to, to_tag_new_random ());
	}
    }

  i = from_clone (request->from, &(response->from));
  if (i != 0)
    goto mcubr_error_1;

  pos = 0;
  while (!list_eol (request->vias, pos))
    {
      via_t *via;
      via_t *via2;

      via = (via_t *) list_get (request->vias, pos);
      i = via_clone (via, &via2);
      if (i != -0)
	goto mcubr_error_1;
      list_add (response->vias, via2, -1);
      pos++;
    }

  i = call_id_clone (request->call_id, &(response->call_id));
  if (i != 0)
    goto mcubr_error_1;
  i = cseq_clone (request->cseq, &(response->cseq));
  if (i != 0)
    goto mcubr_error_1;

  *dest = response;
  return 0;

mcubr_error_1:
  msg_free (response);
  sfree (response);
  return -1;
}

int
osip_msg_sfp_build_response_to_forward (sip_t ** dest, sip_t * response)
{
  sip_t *fwd;
  via_t *via;
  int i;

  *dest = NULL;
  i = msg_clone (response, &fwd);
  if (i != 0)
    return -1;

  /* remove top via... and send */
  via = list_get (fwd->vias, 0);
  if (via == NULL)		/* remote error: no via!! */
    {
      msg_free (fwd);
      sfree (fwd);
      return -1;
    }
  list_remove (fwd->vias, 0);
  via_free (via);
  sfree (via);

  *dest = fwd;
  return 0;
}

int
osip_msg_sfp_build_request_to_forward (psp_req_t * req, sip_t ** dest,
				       url_t * rquri,   char *path,
				       sip_t * request, int stayonpath)
{
  generic_param_t *branch_param;
  via_t *via;
  char *branch;
  int i;
  char number[3];

  i =
    osip_msg_default_build_request_to_forward (dest, rquri, path, request,
					       stayonpath);
  if (i != 0)
    return -1;

  /* ACK don't need a specific Via branch (the one just generated is fine!).
     This method will only be called when forwarding ACK for 2xx! (in the other
     cases, the transaction layer will handle it alone. */

  /* calculate the branch number */
  if (!MSG_IS_ACK (request))
    {
      /* fix the via... so we'll match the modified Via with the slp_branch. */
      branch = psp_core_sfp_generate_branch_for_request (request);
      if (branch == NULL)
	{
	  msg_free (*dest);
	  sfree (*dest);
	  *dest = NULL;
	  return -1;		/* there are some cases where we can't
				   build the branch? */
	}
      if (req->rsp_ctx->ind > 50)
	{
	  msg_free (*dest);
	  sfree (*dest);
	  *dest = NULL;
	  sfree (branch);
	  return -1;
	}

      sprintf (number, "%i", req->rsp_ctx->ind);
      req->rsp_ctx->ind++;

      branch[40] = '.';
      branch[41] = number[0];
      branch[42] = number[1];
      branch[43] = '\0';

      /* get the Top Via and fix the branch */
      msg_getvia (*dest, 0, &via);
      if (via == NULL)
	{
	  msg_free (*dest);
	  sfree (*dest);
	  *dest = NULL;
	  sfree (branch);
	  return -1;
	}
      via_param_getbyname (via, "branch", &branch_param);
      if (branch_param == NULL)
	{			/* BUG? */
	  msg_free (*dest);
	  sfree (*dest);
	  *dest = NULL;
	  sfree (branch);
	  return -1;
	}
      sfree (branch_param->gvalue);
      branch_param->gvalue = branch;
    }

  return 0;
}

/* Add a header to a SIP message at the top of the list.    */
/* This method is taken from libosip-0.9.3 so partysip
   is still able to link with older osip version. */
int
_msg_settopheader (sip_t * sip, char *hname, char *hvalue)
{
  header_t *h;
  int i;

  if (hname == NULL)
    return -1;

  i = header_init (&h);
  if (i != 0)
    return -1;

  h->hname = (char *) smalloc (strlen (hname) + 1);

  if (h->hname == NULL)
    {
      header_free (h);
      sfree (h);
      return -1;
    }
  sstrncpy (h->hname, hname, strlen (hname));
  sclrspace (h->hname);

  if (hvalue != NULL)
    {				/* some headers can be null ("subject:") */
      h->hvalue = (char *) smalloc (strlen (hvalue) + 1);
      if (h->hvalue == NULL)
	{
	  header_free (h);
	  sfree (h);
	  return -1;
	}
      sstrncpy (h->hvalue, hvalue, strlen (hvalue));
      sclrspace (h->hvalue);
    }
  else
    h->hvalue = NULL;
#ifdef USE_TMP_BUFFER
  sip->message_property = 2;
#endif
  list_add (sip->headers, h, 0);
  return 0;			/* ok */
}

/* keep this one ! */
int
osip_msg_default_build_request_to_forward (sip_t ** dest, url_t * rquri,
					   char *path,    sip_t * request,
					   int stayonpath)
{
  header_t *maxfwd;
  via_t *via;
  route_t *route;
  route_t *route2;
  sip_t *fwd;
  char *branch;
  char *serverip;
  char *serverport;
  int i;

  *dest = NULL;
  i = msg_clone (request, &fwd);
  if (i != 0)
    return -1;


  /* TODO from rfc3261:
     if the URI contains unknown parameters, they MUST be removed.
   */

  /* add a via header... */
  branch = psp_core_default_generate_branch_for_request (request);
  if (branch == NULL)
    goto pclbr_error_1;		/* there are some cases where we can't
				   build the branch? */

  /* Oupss! BUG
     serverip = psp_config_get_element ("serverip");
  */
  serverip=NULL;
  if (core->ext_ip!=NULL)
    serverip = core->ext_ip;
  if (serverip==NULL)
    serverip = core->serverip[0];

  if (serverip == NULL)
     {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "Bad configuration\n"));
      goto pclbr_error_1;
    }
  serverport = psp_config_get_element ("serverport_udp");

  /*        8.   Add a Via header field value   */

  i = via_init (&via);
  if (i != 0)
    goto pclbr_error_1;
  i = via_set_branch (via, branch);
  if (i != 0)
    {
      via_free (via);
      sfree (via);
      goto pclbr_error_2;
    }
  via_setversion (via, sgetcopy ("2.0"));
  via_sethost (via, sgetcopy (serverip));
  /** SUPPORT ONLY UDP BY NOW!! **/
  if (serverport != NULL)
    via_setport (via, sgetcopy (serverport));
  via_setprotocol (via, sgetcopy ("UDP"));

  list_add (fwd->vias, via, 0);

  /*        3.   Update the Max-Forwards header field */
  msg_getmax_forward (fwd, 0, &maxfwd);
  if (maxfwd != NULL)
    {
      char *tmp;
      int maxf;

      if (maxfwd->hvalue != NULL)
	{
	  maxf = satoi (maxfwd->hvalue);
	  maxf--;
	  tmp = (char *) smalloc (strlen (maxfwd->hvalue) + 2);
	  if (tmp == NULL)
	    goto pclbr_error_2;
	  sfree (maxfwd->hvalue);
	  sprintf (tmp, "%i", maxf);
	  maxfwd->hvalue = tmp;
	}
    }

  /* route pre-processing:
     "If the first value in the Route header field indicates this proxy,
     the proxy MUST remove that value from the request." */
  msg_getroute (fwd, 0, &route);
  if (route != NULL)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO2, NULL,
		   "Checking if we are responsible for '%s'\n",
		   route->url->host));
      i = psp_core_is_responsible_for_this_route (route->url);
      if (i == 0)		/* yes, it is */
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO4, NULL,
		       "I detect a route inserted by me, I remove it.\n"));
	  list_remove (fwd->routes, 0);
	  route_free (route);
	  sfree (route);
	}
    }
  /*        4.   Optionally add a Record-route header field value */
  if (stayonpath && MSG_IS_INVITE (request))
    {				/* AMD: 14/10/2002 This information is usefull only for INVITE */
      record_route_t *r_route;
      char *tmp;
      url_t *url_of_proxy;

      i = record_route_init (&r_route);
      if (i != 0)
	goto pclbr_error_2;
      i = url_init (&url_of_proxy);
      if (i != 0)
	{
	  record_route_free (r_route);
	  sfree (r_route);
	  goto pclbr_error_2;
	}
      url_sethost (url_of_proxy,
		   sgetcopy (psp_core_get_ip_of_local_proxy ()));
      tmp = (char *) smalloc (9);
      sprintf (tmp, "%i", psp_core_get_port_of_local_proxy ());
      url_setport (url_of_proxy, tmp);
      url_uparam_add (url_of_proxy, sgetcopy ("lr"), NULL);
      url_uparam_add (url_of_proxy, sgetcopy ("psp"),
		      sgetcopy (psp_config_get_element ("magicstring")));
      record_route_seturl (r_route, url_of_proxy);
      /* insert above all other record-route */
      list_add (fwd->record_routes, r_route, 0);
    }

  /*        9.   Add a Content-Length header field if necessary */
  if (fwd->contentlength == NULL)
    {
      if (list_size (fwd->bodies) == 0)
	msg_setcontent_length (fwd, "0");
      /* else should be refused before... but I prefer to let
         other deal with that */
    }

  /* several annoying cases:
     No locations were found but a location comes from route processing: */
  msg_getroute (request, 0, &route);
  msg_getroute (request, 1, &route2);
  if (route2 != NULL)
    {
    }				/* We are sure to use the record-route mechanism and message
				   is already correct. */
  else if (route == NULL ||
	   0 == psp_core_is_responsible_for_this_route (route->url))
    {				/* The message might not be already correct if the route is not me */
      if (rquri != NULL)
	{
	  url_free (fwd->strtline->rquri);
	  sfree (fwd->strtline->rquri);
	  fwd->strtline->rquri = NULL;	/* mandatory, bcz the next call can fail
					   and address MUST never be NULL if not
					   allocated... */
	  msg_seturi (fwd, rquri);
	}
    }

  if (path!=NULL)
    {
      char *beg;
      char *end;
      /* A location contains a set of Path headers. Those headers MUST
	 be set appended as "Route" headers to the new Request */
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO4, NULL,
		   "osip_msg module: build a request with a Pre-Route set.\n"));
      beg = path;
      end = strchr(beg, ','); /* hope you don't have a comma in your path :< */
      if (end==NULL)
	{
	   /* Where should we append this route? Top or Bottom?
	      (in fact, we won't probably get any other route header there.) */
	  msg_setroute(fwd, path);
	}
      else
	{
	  /* TODO: please support insertion of several Path header :< */
	  /* TODO */
	}
    }

  /*  handle rfc3327:
      Extension Header Field for Registering Non-Adjacent Contacts */
  if (core->rfc3327==1 && MSG_IS_REGISTER(fwd)) /* option is active */
    {
      /* add a path field pointing to myself. (so I'll be on the route
	 of incoming requests coming from the registrar.) */
      record_route_t *r_route;
      char *tmp;
      url_t *url_of_proxy;

      i = record_route_init (&r_route);
      if (i != 0)
	goto pclbr_error_2;
      i = url_init (&url_of_proxy);
      if (i != 0)
	{
	  record_route_free (r_route);
	  sfree (r_route);
	  goto pclbr_error_2;
	}
      url_sethost (url_of_proxy,
		   sgetcopy (psp_core_get_ip_of_local_proxy ()));
      tmp = (char *) smalloc (9);
      sprintf (tmp, "%i", psp_core_get_port_of_local_proxy ());
      url_setport (url_of_proxy, tmp);
      url_uparam_add (url_of_proxy, sgetcopy ("lr"), NULL);
      url_uparam_add (url_of_proxy, sgetcopy ("psp"),
		      sgetcopy (psp_config_get_element ("magicstring")));
      record_route_seturl (r_route, url_of_proxy);

      /* get string for new 'Path' header: */
      i = record_route_2char(r_route, &tmp);
      record_route_free (r_route);
      sfree (r_route);
      if (i != 0)
	{
	  goto pclbr_error_2;
	}
      /* insert above all other Path header */
      _msg_settopheader(fwd, "Path", tmp);
      sfree(tmp);
    }


  *dest = fwd;
  return 0;

pclbr_error_2:
  msg_free (fwd);
  sfree (fwd);
  return -1;
pclbr_error_1:
  msg_free (fwd);
  sfree (fwd);
  sfree (branch);
  return -1;
}


int
osip_msg_modify_ack_to_be_forwarded (sip_t * ack, int stayonpath)
{
  header_t *maxfwd;
  via_t *via;
  route_t *route;
  char *branch;
  char *serverip;
  char *serverport;
  int i;

  /* will it be unique? To be verified. */
  branch = psp_core_default_generate_branch_for_request (ack);
  if (branch == NULL)
    return -1;			/* there are some cases where we can't
				   build the branch? */

  /* Oupss! BUG
     serverip = psp_config_get_element ("serverip");
  */
  serverip=NULL;
  if (core->ext_ip!=NULL)
    serverip = core->ext_ip;
  if (serverip==NULL)
    serverip = core->serverip[0];

  if (serverip == NULL)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "Bad configuration\n"));
      goto pclbr_error_1;
    }
  serverport = psp_config_get_element ("serverport_udp");

  /*        8.   Add a Via header field value   */

  i = via_init (&via);
  if (i != 0)
    goto pclbr_error_1;
  i = via_set_branch (via, branch);
  if (i != 0)
    {
      via_free (via);
      sfree (via);
      goto pclbr_error_2;
    }
  via_setversion (via, sgetcopy ("2.0"));
  via_sethost (via, sgetcopy (serverip));
  /** SUPPORT ONLY UDP BY NOW!! **/
  if (serverport != NULL)
    via_setport (via, sgetcopy (serverport));
  via_setprotocol (via, sgetcopy ("UDP"));

  list_add (ack->vias, via, 0);

  /*        3.   Update the Max-Forwards header field */
  msg_getmax_forward (ack, 0, &maxfwd);
  if (maxfwd != NULL)
    {
      char *tmp;
      int maxf;

      if (maxfwd->hvalue != NULL)
	{
	  maxf = satoi (maxfwd->hvalue);
	  maxf--;
	  tmp = (char *) smalloc (strlen (maxfwd->hvalue) + 2);
	  if (tmp == NULL)
	    goto pclbr_error_2;
	  sfree (maxfwd->hvalue);
	  sprintf (tmp, "%i", maxf);
	  maxfwd->hvalue = tmp;
	}
    }

  /* route pre-processing:
     "If the first value in the Route header field indicates this proxy,
     the proxy MUST remove that value from the request." */
  msg_getroute (ack, 0, &route);
  if (route != NULL)
    {
      i = psp_core_is_responsible_for_this_route (route->url);
      if (i == 0)		/* yes, it is */
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO4, NULL,
		       "I detect a route inserted by me, I remove it.\n"));
	  list_remove (ack->routes, 0);
	  route_free (route);
	  sfree (route);
	  /* in this case, we have previously requested to stay on the path.
	     We don't want to change that decision now, so we override
	     the possible value of stayonpath! */
	  stayonpath = 1;
	}
    }
  /*        4.   Optionally add a Record-route header field value */
  /* AMD: 14/10/2002: ACK don't need record-route info. */

  /*        9.   Add a Content-Length header field if necessary */
  if (ack->contentlength == NULL)
    {
      if (list_size (ack->bodies) == 0)
	msg_setcontent_length (ack, "0");
      /* else should be refused before... but I prefer to let
         other deal with that */
    }

pclbr_error_2:
  return -1;
pclbr_error_1:
  sfree (branch);
  return -1;
}


int
osip_msg_build_cancel (sip_t ** dest, sip_t * request_cancelled)
{
  int i;
  sip_t *request;

  i = msg_init (&request);
  if (i != 0)
    return -1;

  /* prepare the request-line */
  request->strtline->sipmethod = sgetcopy ("CANCEL");
  request->strtline->sipversion = sgetcopy ("SIP/2.0");
  request->strtline->statuscode = NULL;
  request->strtline->reasonphrase = NULL;

  i =
    url_clone (request_cancelled->strtline->rquri,
	       &(request->strtline->rquri));
  if (i != 0)
    goto gc_error_1;

  i = to_clone (request_cancelled->to, &(request->to));
  if (i != 0)
    goto gc_error_1;
  i = from_clone (request_cancelled->from, &(request->from));
  if (i != 0)
    goto gc_error_1;

  /* set the cseq and call_id header */
  i = call_id_clone (request_cancelled->call_id, &(request->call_id));
  if (i != 0)
    goto gc_error_1;
  i = cseq_clone (request_cancelled->cseq, &(request->cseq));
  if (i != 0)
    goto gc_error_1;
  sfree (request->cseq->method);
  request->cseq->method = sgetcopy ("CANCEL");

  /* copy ONLY the top most Via Field (this method is also used by proxy) */
  {
    via_t *via;
    via_t *via2;

    i = msg_getvia (request_cancelled, 0, &via);
    if (i != 0)
      goto gc_error_1;
    i = via_clone (via, &via2);
    if (i != 0)
      goto gc_error_1;
    list_add (request->vias, via2, -1);
  }

  msg_setcontent_length (request, "0");
  msg_setmax_forward (request, "70");	/* a UA should start a request with 70 */
  msg_setuser_agent (request, "oSIP-ua/0.8.1");

  *dest = request;
  return 0;

gc_error_1:
  msg_free (request);
  sfree (request);
  *dest = NULL;
  return -1;
}
