/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>

extern psp_core_t *core;

/************ PLUGIN REGISTRATION API *******************/


PPL_DECLARE (int)
psp_core_load_tlp_plugin (tlp_plugin_t ** p, psp_plugin_t * psp, int flag)
{
  tlp_plugin_t *plug;
  int i;

  *p = NULL;
  i = tlp_plugin_init (&plug, psp, flag);
  if (i != 0)
    return -1;
  *p = plug;
  psp_core_add_tlp_plugin (plug);
  return 0;
}

PPL_DECLARE (int) psp_core_add_tlp_plugin (tlp_plugin_t * plug)
{
  ADD_ELEMENT (core->tlp->tlp_plugins, plug);
  return 0;
}

PPL_DECLARE (int)
psp_core_load_imp_plugin (imp_plugin_t ** p, psp_plugin_t * psp)
{
  imp_plugin_t *plug;
  int i;

  *p = NULL;
  i = imp_plugin_init (&plug, psp);
  if (i != 0)
    return -1;
  *p = plug;
  psp_core_add_imp_plugin (plug);
  return 0;
}

PPL_DECLARE (int) psp_core_add_imp_plugin (imp_plugin_t * plug)
{
  ADD_ELEMENT (core->imp->imp_plugins, plug);
  return 0;
}

PPL_DECLARE (int)
psp_core_load_uap_plugin (uap_plugin_t ** p, psp_plugin_t * psp)
{
  uap_plugin_t *plug;
  int i;

  *p = NULL;
  i = uap_plugin_init (&plug, psp);
  if (i != 0)
    return -1;
  *p = plug;
  psp_core_add_uap_plugin (plug);
  return 0;
}

PPL_DECLARE (int) psp_core_add_uap_plugin (uap_plugin_t * plug)
{
  ADD_ELEMENT (core->uap->uap_plugins, plug);
  return 0;
}

PPL_DECLARE (int)
psp_core_load_sfp_plugin (sfp_plugin_t ** p, psp_plugin_t * psp)
{
  sfp_plugin_t *plug;
  int i;

  *p = NULL;
  i = sfp_plugin_init (&plug, psp);
  if (i != 0)
    return -1;
  *p = plug;
  psp_core_add_sfp_plugin (plug);
  return 0;
}

PPL_DECLARE (int) psp_core_add_sfp_plugin (sfp_plugin_t * plug)
{
  ADD_ELEMENT (core->sfp->sfp_plugins, plug);
  return 0;
}

/************ HOOK REGISTRATION API *********************/

int
psp_core_add_imp_hook (imp_func_tab_t * tab, imp_func_t * fn, int hookflg)
{
  switch (hookflg)
    {
    case PSP_HOOK_REALLY_FIRST:
      APPEND_ELEMENT (imp_func_t, tab->func_hook_really_first, fn);
      break;
    case PSP_HOOK_FIRST:
      APPEND_ELEMENT (imp_func_t, tab->func_hook_first, fn);
      break;
    case PSP_HOOK_LAST:
      APPEND_ELEMENT (imp_func_t, tab->func_hook_last, fn);
      break;
    case PSP_HOOK_REALLY_LAST:
      APPEND_ELEMENT (imp_func_t, tab->func_hook_last, fn);
      break;
    case PSP_HOOK_MIDDLE:
    default:
      APPEND_ELEMENT (imp_func_t, tab->func_hook_middle, fn);
      break;
    }
  return 0;
}


PPL_DECLARE (int) psp_core_add_imp_invite_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_invites, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_ack_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_acks, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_imp_register_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_registers, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_bye_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_byes, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_options_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_optionss, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_info_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_infos, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_cancel_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_cancels, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_notify_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_notifys, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_imp_subscribe_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_subscribes, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_imp_unknown_hook (imp_func_t * fn, int hookflg)
{
  return psp_core_add_imp_hook (core->imp->func_unknowns, fn, hookflg);
}

int
psp_core_remove_imp_all_hook ()
{
  /* these elements are just pointers to static reference. We don't need
     to delete them... */
  core->imp->func_invites = NULL;
  core->imp->func_acks = NULL;
  core->imp->func_registers = NULL;
  core->imp->func_byes = NULL;
  core->imp->func_optionss = NULL;
  core->imp->func_infos = NULL;
  core->imp->func_cancels = NULL;
  core->imp->func_notifys = NULL;
  core->imp->func_subscribes = NULL;
  core->imp->func_unknowns = NULL;
  return 0;
}

/***UAP***/

int
psp_core_add_uap_hook (uap_func_tab_t * tab, uap_func_t * fn, int hookflg)
{
  switch (hookflg)
    {
    case PSP_HOOK_REALLY_FIRST:
      APPEND_ELEMENT (uap_func_t, tab->func_hook_really_first, fn);
      break;
    case PSP_HOOK_FIRST:
      APPEND_ELEMENT (uap_func_t, tab->func_hook_first, fn);
      break;
    case PSP_HOOK_LAST:
      APPEND_ELEMENT (uap_func_t, tab->func_hook_last, fn);
      break;
    case PSP_HOOK_REALLY_LAST:
      APPEND_ELEMENT (uap_func_t, tab->func_hook_really_last, fn);
      break;
    case PSP_HOOK_MIDDLE:
    default:
      APPEND_ELEMENT (uap_func_t, tab->func_hook_middle, fn);
      break;
    }
  return 0;
}

PPL_DECLARE (int) psp_core_add_uap_snd_1xx_hook (uap_func_t * fn, int hookflg)
{
  return psp_core_add_uap_hook (core->uap->func_1xxs, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_uap_snd_2xx_hook (uap_func_t * fn, int hookflg)
{
  return psp_core_add_uap_hook (core->uap->func_2xxs, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_uap_snd_3xx_hook (uap_func_t * fn, int hookflg)
{
  return psp_core_add_uap_hook (core->uap->func_3xxs, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_uap_snd_4xx_hook (uap_func_t * fn, int hookflg)
{
  return psp_core_add_uap_hook (core->uap->func_4xxs, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_uap_snd_5xx_hook (uap_func_t * fn, int hookflg)
{
  return psp_core_add_uap_hook (core->uap->func_5xxs, fn, hookflg);
}

PPL_DECLARE (int) psp_core_add_uap_snd_6xx_hook (uap_func_t * fn, int hookflg)
{
  return psp_core_add_uap_hook (core->uap->func_6xxs, fn, hookflg);
}

int
psp_core_remove_uap_all_hook ()
{
  core->uap->func_1xxs = NULL;
  core->uap->func_2xxs = NULL;
  core->uap->func_3xxs = NULL;
  core->uap->func_4xxs = NULL;
  core->uap->func_5xxs = NULL;
  core->uap->func_6xxs = NULL;
  return 0;
}


/***SFP***/
int
psp_core_add_sfp_fwd_hook (sfp_fwd_func_tab_t * tab, sfp_fwd_func_t * fn,
			   int hookflg)
{
  switch (hookflg)
    {
    case PSP_HOOK_REALLY_FIRST:
      APPEND_ELEMENT (sfp_fwd_func_t, tab->func_hook_really_first, fn);
      break;
    case PSP_HOOK_FIRST:
      APPEND_ELEMENT (sfp_fwd_func_t, tab->func_hook_first, fn);
      break;
    case PSP_HOOK_LAST:
      APPEND_ELEMENT (sfp_fwd_func_t, tab->func_hook_last, fn);
      break;
    case PSP_HOOK_REALLY_LAST:
      APPEND_ELEMENT (sfp_fwd_func_t, tab->func_hook_really_last, fn);
      break;
    case PSP_HOOK_MIDDLE:
    default:
      APPEND_ELEMENT (sfp_fwd_func_t, tab->func_hook_middle, fn);
      break;
    }
  return 0;
}

int
psp_core_add_sfp_rcv_hook (sfp_rcv_func_tab_t * tab, sfp_rcv_func_t * fn,
			   int hookflg)
{
  switch (hookflg)
    {
    case PSP_HOOK_REALLY_FIRST:
      APPEND_ELEMENT (sfp_rcv_func_t, tab->func_hook_really_first, fn);
      break;
    case PSP_HOOK_FIRST:
      APPEND_ELEMENT (sfp_rcv_func_t, tab->func_hook_first, fn);
      break;
    case PSP_HOOK_LAST:
      APPEND_ELEMENT (sfp_rcv_func_t, tab->func_hook_last, fn);
      break;
    case PSP_HOOK_REALLY_LAST:
      APPEND_ELEMENT (sfp_rcv_func_t, tab->func_hook_really_last, fn);
      break;
    case PSP_HOOK_MIDDLE:
    default:
      APPEND_ELEMENT (sfp_rcv_func_t, tab->func_hook_middle, fn);
      break;
    }
  return 0;
}

int
psp_core_add_sfp_snd_hook (sfp_snd_func_tab_t * tab, sfp_snd_func_t * fn,
			   int hookflg)
{
  switch (hookflg)
    {
    case PSP_HOOK_REALLY_FIRST:
      APPEND_ELEMENT (sfp_snd_func_t, tab->func_hook_really_first, fn);
      break;
    case PSP_HOOK_FIRST:
      APPEND_ELEMENT (sfp_snd_func_t, tab->func_hook_first, fn);
      break;
    case PSP_HOOK_LAST:
      APPEND_ELEMENT (sfp_snd_func_t, tab->func_hook_last, fn);
      break;
    case PSP_HOOK_REALLY_LAST:
      APPEND_ELEMENT (sfp_snd_func_t, tab->func_hook_really_last, fn);
      break;
    case PSP_HOOK_MIDDLE:
    default:
      APPEND_ELEMENT (sfp_snd_func_t, tab->func_hook_middle, fn);
      break;
    }
  return 0;
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_invite_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_invites, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_ack_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_acks, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_register_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_registers, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_bye_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_byes, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_options_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_optionss, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_info_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_infos, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_cancel_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_cancels, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_notify_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_notifys, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_subscribe_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_subscribes, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_fwd_unknown_hook (sfp_fwd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_fwd_hook (core->sfp->fwd_unknowns, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_rcv_1xx_hook (sfp_rcv_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_rcv_hook (core->sfp->rcv_1xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_rcv_2xx_hook (sfp_rcv_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_rcv_hook (core->sfp->rcv_2xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_rcv_3xx_hook (sfp_rcv_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_rcv_hook (core->sfp->rcv_3xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_rcv_4xx_hook (sfp_rcv_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_rcv_hook (core->sfp->rcv_4xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_rcv_5xx_hook (sfp_rcv_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_rcv_hook (core->sfp->rcv_5xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_rcv_6xx_hook (sfp_rcv_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_rcv_hook (core->sfp->rcv_6xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_snd_1xx_hook (sfp_snd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_snd_hook (core->sfp->snd_1xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_snd_2xx_hook (sfp_snd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_snd_hook (core->sfp->snd_2xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_snd_3xx_hook (sfp_snd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_snd_hook (core->sfp->snd_3xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_snd_4xx_hook (sfp_snd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_snd_hook (core->sfp->snd_4xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_snd_5xx_hook (sfp_snd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_snd_hook (core->sfp->snd_5xxs, fn, hookflg);
}

PPL_DECLARE (int)
psp_core_add_sfp_snd_6xx_hook (sfp_snd_func_t * fn, int hookflg)
{
  return psp_core_add_sfp_snd_hook (core->sfp->snd_6xxs, fn, hookflg);
}

int
psp_core_remove_sfp_all_hook ()
{
  core->sfp->fwd_invites = NULL;
  core->sfp->fwd_acks = NULL;
  core->sfp->fwd_registers = NULL;
  core->sfp->fwd_byes = NULL;
  core->sfp->fwd_optionss = NULL;
  core->sfp->fwd_infos = NULL;
  core->sfp->fwd_cancels = NULL;
  core->sfp->fwd_notifys = NULL;
  core->sfp->fwd_subscribes = NULL;
  core->sfp->fwd_unknowns = NULL;

  core->sfp->rcv_1xxs = NULL;
  core->sfp->rcv_2xxs = NULL;
  core->sfp->rcv_3xxs = NULL;
  core->sfp->rcv_4xxs = NULL;
  core->sfp->rcv_5xxs = NULL;
  core->sfp->rcv_6xxs = NULL;

  core->sfp->snd_1xxs = NULL;
  core->sfp->snd_2xxs = NULL;
  core->sfp->snd_3xxs = NULL;
  core->sfp->snd_4xxs = NULL;
  core->sfp->snd_5xxs = NULL;
  core->sfp->snd_6xxs = NULL;

  return 0;

}
