/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>

extern psp_core_t *core;

void *
osip_timers_thread (psp_osip_t * psp_osip)
{

  fd_set osip_fdset;
  struct timeval tv;

  while (1)
    {

      int s;
      int i;

      tv.tv_sec = 0;
      tv.tv_usec = psp_osip->delay * 1000;

      FD_ZERO (&osip_fdset);
      s = ppl_pipe_get_read_descr (core->psp_osip->wakeup);
      if (s <= 1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_BUG, NULL,
		       "No wakeup value in module osip!\n"));
	  return NULL;
	}
      FD_SET (s, &osip_fdset);

      i = select (s + 1, &osip_fdset, NULL, NULL, &tv);

      if (FD_ISSET (s, &osip_fdset))
	{
	  char tmp[51];

	  i = ppl_pipe_read (core->psp_osip->wakeup, tmp, 50);
	  if (i == 1 && tmp[0] == 'q')
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_INFO2, NULL,
			   "osip module: Exiting!\n"));
	      return 0;
	    }
	  else
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_INFO2, NULL,
			   "osip module: wake up!\n"));
	    }
	}

      /*      susleep (); */
      osip_timers_ict_execute (psp_osip->osip);
      osip_timers_ist_execute (psp_osip->osip);
      osip_timers_nict_execute (psp_osip->osip);
      osip_timers_nist_execute (psp_osip->osip);

      psp_core_lock ();
      osip_ict_execute (psp_osip->osip);
      osip_nict_execute (psp_osip->osip);
      osip_ist_execute (psp_osip->osip);
      osip_nist_execute (psp_osip->osip);
      psp_core_unlock ();

      if (psp_osip->exit_flag == 1)
	return NULL;
    }
  return 0;
}

int
psp_osip_init (psp_osip_t ** psp_osip, int delay)
{
  osip_t *osip;
  int i;

  /* initialize osip */
  if (-1 == osip_global_init ())
    return -1;			/* mutex is not initialised properly */

  i = osip_init (&osip);
  if (i != 0)
    return -1;

  osip_setcb_send_message (osip, &psp_core_cb_snd_message);

  osip_setcb_ict_kill_transaction (osip, &psp_core_cb_ict_kill_transaction);
  osip_setcb_ist_kill_transaction (osip, &psp_core_cb_ist_kill_transaction);
  osip_setcb_nict_kill_transaction (osip, &psp_core_cb_nict_kill_transaction);
  osip_setcb_nist_kill_transaction (osip, &psp_core_cb_nist_kill_transaction);

  osip_setcb_ict_2xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_ict_3456xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_ict_invite_sent2 (osip, &psp_core_cb_sndreq_retransmission);
  osip_setcb_ist_2xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_ist_3456xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_ist_invite_received2 (osip, &psp_core_cb_rcvreq_retransmission);
  osip_setcb_nict_2xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_nict_3456xx_received2 (osip,
				    &psp_core_cb_rcvresp_retransmission);
  osip_setcb_nict_request_sent2 (osip, &psp_core_cb_sndreq_retransmission);
  osip_setcb_nist_2xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_nist_3456xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_nist_request_received2 (osip,
				     &psp_core_cb_rcvreq_retransmission);

  /*  osip_setcb_killtransaction(osip,&psp_core_cb_killtransaction);
     osip_setcb_endoftransaction(osip,&psp_core_cb_endoftransaction); */

  osip_setcb_ict_transport_error (osip, &psp_core_cb_transport_error);
  osip_setcb_ist_transport_error (osip, &psp_core_cb_transport_error);
  osip_setcb_nict_transport_error (osip, &psp_core_cb_transport_error);
  osip_setcb_nist_transport_error (osip, &psp_core_cb_transport_error);

  osip_setcb_ict_invite_sent (osip, &psp_core_cb_sndinvite);
  osip_setcb_ict_ack_sent (osip, &psp_core_cb_sndack);
  osip_setcb_nict_register_sent (osip, &psp_core_cb_sndregister);
  osip_setcb_nict_bye_sent (osip, &psp_core_cb_sndbye);
  osip_setcb_nict_cancel_sent (osip, &psp_core_cb_sndcancel);
  osip_setcb_nict_info_sent (osip, &psp_core_cb_sndinfo);
  osip_setcb_nict_options_sent (osip, &psp_core_cb_sndoptions);
  osip_setcb_nict_subscribe_sent (osip, &psp_core_cb_sndoptions);
  osip_setcb_nict_notify_sent (osip, &psp_core_cb_sndoptions);
  /*  osip_setcb_nict_sndprack   (osip,&psp_core_cb_sndprack); */
  osip_setcb_nict_unknown_sent (osip, &psp_core_cb_sndunkrequest);

  osip_setcb_ict_1xx_received (osip, &psp_core_cb_rcv1xx);
  osip_setcb_ict_2xx_received (osip, &psp_core_cb_rcv2xx);
  osip_setcb_ict_3xx_received (osip, &psp_core_cb_rcv3xx);
  osip_setcb_ict_4xx_received (osip, &psp_core_cb_rcv4xx);
  osip_setcb_ict_5xx_received (osip, &psp_core_cb_rcv5xx);
  osip_setcb_ict_6xx_received (osip, &psp_core_cb_rcv6xx);

  osip_setcb_ist_1xx_sent (osip, &psp_core_cb_snd1xx);
  osip_setcb_ist_2xx_sent (osip, &psp_core_cb_snd2xx);
  osip_setcb_ist_3xx_sent (osip, &psp_core_cb_snd3xx);
  osip_setcb_ist_4xx_sent (osip, &psp_core_cb_snd4xx);
  osip_setcb_ist_5xx_sent (osip, &psp_core_cb_snd5xx);
  osip_setcb_ist_6xx_sent (osip, &psp_core_cb_snd6xx);

  osip_setcb_nict_1xx_received (osip, &psp_core_cb_rcv1xx);
  osip_setcb_nict_2xx_received (osip, &psp_core_cb_rcv2xx);
  osip_setcb_nict_3xx_received (osip, &psp_core_cb_rcv3xx);
  osip_setcb_nict_4xx_received (osip, &psp_core_cb_rcv4xx);
  osip_setcb_nict_5xx_received (osip, &psp_core_cb_rcv5xx);
  osip_setcb_nict_6xx_received (osip, &psp_core_cb_rcv6xx);

  osip_setcb_nist_1xx_sent (osip, &psp_core_cb_snd1xx);
  osip_setcb_nist_2xx_sent (osip, &psp_core_cb_snd2xx);
  osip_setcb_nist_3xx_sent (osip, &psp_core_cb_snd3xx);
  osip_setcb_nist_4xx_sent (osip, &psp_core_cb_snd4xx);
  osip_setcb_nist_5xx_sent (osip, &psp_core_cb_snd5xx);
  osip_setcb_nist_6xx_sent (osip, &psp_core_cb_snd6xx);

  osip_setcb_ist_invite_received (osip, &psp_core_cb_rcvinvite);
  osip_setcb_ist_ack_received (osip, &psp_core_cb_rcvack);
  osip_setcb_ist_ack_received2 (osip, &psp_core_cb_rcvack2);
  osip_setcb_nist_register_received (osip, &psp_core_cb_rcvregister);
  osip_setcb_nist_bye_received (osip, &psp_core_cb_rcvbye);
  osip_setcb_nist_cancel_received (osip, &psp_core_cb_rcvcancel);
  osip_setcb_nist_info_received (osip, &psp_core_cb_rcvinfo);
  osip_setcb_nist_options_received (osip, &psp_core_cb_rcvoptions);
  osip_setcb_nist_subscribe_received (osip, &psp_core_cb_rcvoptions);
  osip_setcb_nist_notify_received (osip, &psp_core_cb_rcvoptions);
  osip_setcb_nist_unknown_received (osip, &psp_core_cb_rcvunkrequest);

  (*psp_osip) = (psp_osip_t *) smalloc (sizeof (psp_osip_t));
  if ((*psp_osip) == NULL)
    goto poi_error0;

  (*psp_osip)->osip = osip;

  if (100 < delay < 500)
    (*psp_osip)->delay = delay;
  else
    delay = 500;

  (*psp_osip)->exit_flag = 0;	/* 0 not set, 1 set */
  (*psp_osip)->timers = NULL;

  (*psp_osip)->wakeup = ppl_pipe ();
  if ((*psp_osip)->wakeup == NULL)
    goto poi_error0;

  (*psp_osip)->timers =
    sthread_create (20000, NULL, (void *(*)(void *)) osip_timers_thread,
		    (void *) (*psp_osip));

  if ((*psp_osip)->timers == NULL)
    goto poi_error1;

  return 0;

poi_error1:
  (*psp_osip)->exit_flag = 1;
  psp_osip_free (*psp_osip);
  sfree (*psp_osip);
  return -1;
poi_error0:
  osip_free (osip);
  sfree (osip);
  return -1;
}

int
psp_osip_wakeup (psp_osip_t * psp_osip)
{
  int i;
  char q[2] = "w";

  if (psp_osip == NULL)
    return -1;
  i = ppl_pipe_write (psp_osip->wakeup, &q, 1);
  if (i != 1)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "could not write in pipe!\n"));
      perror ("error while writing in pipe");
      return -1;
    }
  return 0;
}

void
_osip_kill_transaction (list_t * transactions)
{
  transaction_t *transaction;

  if (!list_eol (transactions, 0))
    {
      /* some transaction are still used by osip,
         transaction should be released by modules! */
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: _osip_kill_transaction transaction should be released by modules!\n"));
    }

  while (!list_eol (transactions, 0))
    {
      transaction = list_get (transactions, 0);
      list_remove (transactions, 0);
      transaction_free (transaction);
      sfree (transaction);
    }
}

void
psp_osip_free (psp_osip_t * psp_osip)
{
  char q[2] = "q";
  int i;

  if (psp_osip == NULL)
    return;

  psp_osip->exit_flag = 1;	/* this value is checked by the timer thread */
  if (psp_osip->timers != NULL)
    {
      i = ppl_pipe_write (psp_osip->wakeup, &q, 1);
      if (i != 1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_BUG, NULL,
		       "could not write in pipe!\n"));
	  return;
	}
      i = sthread_join (psp_osip->timers);
      if (i != 0)
	printf ("ERROR: can't stop timer thread\n");
      sfree (psp_osip->timers);
      psp_osip->timers = NULL;
    }

  /* now it's time to free all transactions... */
  _osip_kill_transaction (psp_osip->osip->ict_transactions);
  _osip_kill_transaction (psp_osip->osip->nict_transactions);
  _osip_kill_transaction (psp_osip->osip->ist_transactions);
  _osip_kill_transaction (psp_osip->osip->nist_transactions);

  i = ppl_pipe_close (psp_osip->wakeup);
  if (i == -1)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "could not close pipe!\n"));
    }
  psp_osip->wakeup = NULL;

  osip_free (psp_osip->osip);
  sfree (psp_osip->osip);
  osip_global_free ();
}


int
psp_osip_release (psp_osip_t * psp_osip)
{
  char q[2] = "q";
  int i;

  if (psp_osip == NULL)
    return -1;
  psp_osip->exit_flag = 1;
  if (psp_osip->timers != NULL)
    {
      i = ppl_pipe_write (psp_osip->wakeup, &q, 1);
      if (i != 1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "could not write in pipe!\n"));
	  return -1;
	}
      i = sthread_join (psp_osip->timers);
      if (i != 0)
	{
	  psp_osip->timers = NULL;
	  printf ("ERROR: can't stop timer thread\n");
	  return -1;
	}
      else
	sfree (psp_osip->timers);
    }
  psp_osip->timers = NULL;
  return 0;
}

int
psp_osip_set_delay (psp_osip_t * psp_osip, int delay)
{
  if (psp_osip == NULL)
    return -1;
  psp_osip->delay = delay;
  return 0;
}
