/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <psp_plug.h>


int
uap_plugin_init (uap_plugin_t ** plug, psp_plugin_t * psp)
{
  *plug = (uap_plugin_t *) smalloc (sizeof (uap_plugin_t));
  if (*plug == NULL)
    return -1;
  (*plug)->psp_plugin = psp;
  (*plug)->next = NULL;
  (*plug)->parent = NULL;
  return 0;
}

void
uap_plugin_free (uap_plugin_t * plug)
{
  if (plug == NULL)
    return;
  psp_plugin_free (plug->psp_plugin);
}

PPL_DECLARE (int)
uap_func_init (uap_func_t ** func, int (*fn) (psp_req_t *), int plug_id)
{
  *func = (uap_func_t *) smalloc (sizeof (uap_func_t));
  if (*func == NULL)
    return -1;

  (*func)->plug_id = plug_id;
  (*func)->cb_snd_func = fn;
  (*func)->next = NULL;
  (*func)->parent = NULL;
  return 0;
}

PPL_DECLARE (void) uap_func_free (uap_func_t * func)
{
  if (func == NULL)
    return;
}

int
uap_func_tab_init (uap_func_tab_t ** tab)
{
  *tab = (uap_func_tab_t *) smalloc (sizeof (uap_func_tab_t));
  if (*tab == NULL)
    return -1;

  (*tab)->func_hook_really_first = NULL;
  (*tab)->func_hook_first = NULL;
  (*tab)->func_hook_middle = NULL;
  (*tab)->func_hook_last = NULL;
  (*tab)->func_hook_really_last = NULL;

  return 0;
}

void
uap_func_tab_free (uap_func_tab_t * tab)
{
  uap_func_t *f;

  if (tab == NULL)
    return;
  for (f = tab->func_hook_really_first; f != NULL;
       f = tab->func_hook_really_first)
    {
      tab->func_hook_really_first = f->next;
      uap_func_free (f);
      sfree (f);
    }
  for (f = tab->func_hook_first; f != NULL; f = tab->func_hook_first)
    {
      tab->func_hook_first = f->next;
      uap_func_free (f);
      sfree (f);
    }
  for (f = tab->func_hook_middle; f != NULL; f = tab->func_hook_middle)
    {
      tab->func_hook_middle = f->next;
      uap_func_free (f);
      sfree (f);
    }
  for (f = tab->func_hook_last; f != NULL; f = tab->func_hook_last)
    {
      tab->func_hook_last = f->next;
      uap_func_free (f);
      sfree (f);
    }
  for (f = tab->func_hook_really_last; f != NULL;
       f = tab->func_hook_really_last)
    {
      tab->func_hook_really_last = f->next;
      uap_func_free (f);
      sfree (f);
    }
}

int
uap_func_tab_add_hook_really_first (uap_func_tab_t * tab, uap_func_t * fn)
{
  if (tab == NULL)
    return -1;
  ADD_ELEMENT (tab->func_hook_really_first, fn);
  return 0;
}

int
uap_func_tab_add_hook_first (uap_func_tab_t * tab, uap_func_t * fn)
{
  if (tab == NULL)
    return -1;
  ADD_ELEMENT (tab->func_hook_first, fn);
  return 0;
}


int
uap_func_tab_add_hook_middle (uap_func_tab_t * tab, uap_func_t * fn)
{
  if (tab == NULL)
    return -1;
  ADD_ELEMENT (tab->func_hook_middle, fn);
  return 0;
}


int
uap_func_tab_add_hook_really_last (uap_func_tab_t * tab, uap_func_t * fn)
{
  if (tab == NULL)
    return -1;
  ADD_ELEMENT (tab->func_hook_really_last, fn);
  return 0;
}

int
uap_func_tab_add_hook_last (uap_func_tab_t * tab, uap_func_t * fn)
{
  if (tab == NULL)
    return -1;
  ADD_ELEMENT (tab->func_hook_last, fn);
  return 0;
}
